"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.calculateValue = exports.recompute = exports.onSourceComplete = exports.onSourceError = exports.makeColdNode = exports.addChildNode = exports.removeValueObserver = exports.addValueObserver = exports.getQueryValue = exports.nextStoreVersion = exports.nextNodeVersion = exports.nextVersion = exports.getComputationNode = exports.createComputationQuery = exports.createComputationNode = exports.compute = void 0;
const rxjs_1 = require("rxjs");
const utils_1 = require("./utils");
/**
 * Creates a computable query which calculates its values by provided "computation" function and dependencies.
 *
 * Rules of "Computation" function:
 * - it must have no side effects
 * - it recalculates only when specified dependencies are updated
 * - its "formula" may have other sources of values, however they don't trigger updates
 *
 * "Computation" function provides a resolver for using a dependency withing a calculation expression.
 *
 * Dependency can be declared explicitly as an array by the second argument. It has the following advantages:
 * - Faster dependency subscription
 * - Ability to specify extra queries or observables as dependencies
 *
 * A custom value comparator can be specified by "options" object as the second argument.
 * It helps to decide if a new value differs from a previous one in complex cases.
 *
 * @example
 * ```ts
 * const greeting = createStore('Hello');
 * const username = createStore('World');
 *
 * // Dependency are implicitly resolved
 * const message = compute((get) => get(greeting) + ' ' + get(username) + '!');
 *
 * // Dependency declared explicitly
 * const messageUppercase = compute(() => message.get().toUpperCase(), [message]);
 *
 * expect(message.get()).toBe('Hello World!');
 * expect(messageUppercase.get()).toBe('HELLO WORLD!');
 * ```
 */
const compute = (computation, dependenciesOrOptions) => {
    const options = dependenciesOrOptions
        ? Array.isArray(dependenciesOrOptions)
            ? { dependencies: dependenciesOrOptions }
            : dependenciesOrOptions
        : undefined;
    const node = createComputationNode(computation, options);
    return createComputationQuery(node);
};
exports.compute = compute;
const NODES = new WeakMap();
/** @internal */
function createComputationNode(computation, options) {
    const dependencies = options === null || options === void 0 ? void 0 : options.dependencies;
    return {
        computation,
        comparator: options === null || options === void 0 ? void 0 : options.comparator,
        hot: false,
        dependencies: dependencies ? [...new Set(dependencies)] : undefined,
        treeObserverCount: 0,
    };
}
exports.createComputationNode = createComputationNode;
/** @internal */
function createComputationQuery(node) {
    const query = {
        get: () => getQueryValue(node),
        value$: new rxjs_1.Observable((observer) => {
            addValueObserver(node, observer);
            return () => removeValueObserver(node, observer);
        }),
    };
    NODES.set(query, node);
    return query;
}
exports.createComputationQuery = createComputationQuery;
/** @internal */
function getComputationNode(query) {
    return NODES.get(query);
}
exports.getComputationNode = getComputationNode;
/// COMPUTATION ENGINE
let NODE_VERSION = 0;
let STORE_VERSION = 0;
const FAST_QUERY_GETTER = (query, selector) => {
    const value = query.get();
    return selector ? selector(value) : value;
};
/** @internal */
function nextVersion(currentValue) {
    return currentValue >= Number.MAX_SAFE_INTEGER ? 0 : currentValue + 1;
}
exports.nextVersion = nextVersion;
/** @internal */
function nextNodeVersion() {
    STORE_VERSION = nextVersion(STORE_VERSION);
    NODE_VERSION = nextVersion(NODE_VERSION);
}
exports.nextNodeVersion = nextNodeVersion;
/** @internal */
function nextStoreVersion() {
    STORE_VERSION = nextVersion(STORE_VERSION);
}
exports.nextStoreVersion = nextStoreVersion;
function getQueryValue(node) {
    var _a;
    if (((_a = node.valueRef) === null || _a === void 0 ? void 0 : _a.version) === STORE_VERSION) {
        return node.valueRef.value;
    }
    return node.computation(FAST_QUERY_GETTER);
}
exports.getQueryValue = getQueryValue;
function addValueObserver(node, observer) {
    if (!node.observers) {
        node.observers = [];
    }
    node.observers.push(observer);
    makeHotNode(node, observer);
    updateTreeObserverCount(node);
}
exports.addValueObserver = addValueObserver;
function removeValueObserver(node, observer) {
    node.observers = (0, utils_1.removeFromArray)(node.observers, observer);
    updateTreeObserverCount(node);
}
exports.removeValueObserver = removeValueObserver;
function getTreeObserverCount(node) {
    var _a;
    const { children, observers } = node;
    let subtreeCount = 0;
    if (children) {
        for (let i = 0; i < children.length; i++) {
            const childNode = children[i];
            subtreeCount += childNode.treeObserverCount;
        }
    }
    return subtreeCount + ((_a = observers === null || observers === void 0 ? void 0 : observers.length) !== null && _a !== void 0 ? _a : 0);
}
function updateTreeObserverCount(node) {
    node.treeObserverCount = getTreeObserverCount(node);
    const parents = node.parents;
    if (parents) {
        for (let i = 0; i < parents.length; i++) {
            const parentNode = parents[i];
            updateTreeObserverCount(parentNode);
        }
    }
    if (node.treeObserverCount === 0) {
        makeColdNode(node);
    }
}
function makeHotNode(node, observer) {
    var _a;
    let dependencies = node.dependencies;
    let valueRef = ((_a = node.valueRef) === null || _a === void 0 ? void 0 : _a.version) === STORE_VERSION ? node.valueRef : undefined;
    if (dependencies) {
        if (observer && !valueRef) {
            valueRef = node.valueRef = calculate(node.computation);
        }
    }
    else {
        const visitedDeps = new Set();
        const next = calculate(node.computation, (query) => visitedDeps.add(query));
        dependencies = node.dependencies = [...visitedDeps];
        if (observer && !valueRef) {
            valueRef = node.valueRef = next;
        }
    }
    if (dependencies.length > 0 && !node.hot) {
        let depObserver;
        let depsSubscriptions = node.depsSubscriptions;
        if (!depsSubscriptions) {
            depsSubscriptions = node.depsSubscriptions = [];
        }
        for (let i = 0; i < dependencies.length; i++) {
            const parentQuery = dependencies[i];
            if (!parentQuery) {
                throw new TypeError('Incorrect dependency');
            }
            const parentNode = NODES.get(parentQuery);
            if (parentNode) {
                addChildNode(parentNode, node);
            }
            else {
                if (!depObserver) {
                    depObserver = {
                        next: () => onSourceChanged(node),
                        error: (error) => onSourceError(node, error),
                        complete: () => onSourceComplete(node),
                    };
                }
                const subscription = parentQuery.value$.subscribe(depObserver);
                depsSubscriptions.push(() => subscription.unsubscribe());
            }
        }
    }
    node.hot = true;
    if (observer && valueRef) {
        observer.next(valueRef.value);
    }
}
function addChildNode(parent, child) {
    if (!parent.children)
        parent.children = [];
    parent.children.push(child);
    if (!child.parents)
        child.parents = [];
    child.parents.push(parent);
    makeHotNode(parent);
}
exports.addChildNode = addChildNode;
function makeColdNode(node) {
    node.hot = false;
    node.treeObserverCount = 0;
    const { depsSubscriptions, parents } = node;
    if (depsSubscriptions) {
        for (let i = 0; i < depsSubscriptions.length; i++) {
            const unsubscribe = depsSubscriptions[i];
            unsubscribe();
        }
    }
    if (parents) {
        for (let i = 0; i < parents.length; i++) {
            const parent = parents[i];
            parent.children = (0, utils_1.removeFromArray)(parent.children, node);
        }
    }
    node.valueRef = undefined;
    node.depsSubscriptions = undefined;
    node.observers = undefined;
    node.parents = undefined;
    node.children = undefined;
}
exports.makeColdNode = makeColdNode;
function onSourceChanged(node) {
    nextNodeVersion();
    recompute(node);
}
function onSourceError(node, error) {
    const { children, observers } = node;
    if (observers) {
        for (let i = 0; i < observers.length; i++) {
            const observer = observers[i];
            observer.error(error);
        }
    }
    if (children) {
        for (let i = 0; i < children.length; i++) {
            const child = children[i];
            onSourceError(child, error);
        }
    }
}
exports.onSourceError = onSourceError;
function onSourceComplete(node) {
    const { children, observers } = node;
    if (observers) {
        for (let i = 0; i < observers.length; i++) {
            const observer = observers[i];
            observer.complete();
        }
    }
    if (children) {
        for (let i = 0; i < children.length; i++) {
            const child = children[i];
            onSourceComplete(child);
        }
    }
}
exports.onSourceComplete = onSourceComplete;
function recompute(node) {
    if (node.version === NODE_VERSION) {
        return;
    }
    node.version = NODE_VERSION;
    if (node.observers && node.observers.length > 0) {
        calculateValue(node);
    }
    else {
        node.valueRef = undefined;
    }
    if (node.treeObserverCount > 0 && node.children) {
        const children = node.children;
        for (let i = 0; i < children.length; i++) {
            const child = children[i];
            recompute(child);
        }
    }
}
exports.recompute = recompute;
function calculateValue(node) {
    var _a;
    const comparator = (_a = node.comparator) !== null && _a !== void 0 ? _a : utils_1.DEFAULT_COMPARATOR;
    const next = calculate(node.computation);
    const isChanged = node.valueRef
        ? isCalculationChanged(comparator, node.valueRef, next)
        : true;
    if (isChanged) {
        node.valueRef = next;
        if (node.observers) {
            const observers = node.observers;
            for (let i = 0; i < observers.length; i++) {
                const observer = observers[i];
                observer.next(next.value);
            }
        }
    }
    if (!isChanged && node.valueRef) {
        node.valueRef.version = STORE_VERSION;
    }
}
exports.calculateValue = calculateValue;
function calculate(computation, visitor) {
    let params;
    const value = computation((query, selector) => {
        let param = query.get();
        if (selector)
            param = selector(param);
        if (!params)
            params = [];
        params.push(param);
        visitor === null || visitor === void 0 ? void 0 : visitor(query);
        return param;
    });
    return { value, params, version: STORE_VERSION };
}
function isCalculationChanged(comparator, a, b) {
    if (comparator(a.value, b.value)) {
        return false;
    }
    return !(a.params && b.params && isArrayEqual(a.params, b.params));
}
function isArrayEqual(a, b) {
    return a.length === b.length && a.every((value, index) => b[index] === value);
}
//# sourceMappingURL=compute.js.map