"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.withStoreUpdates = exports.createStoreUpdates = exports.createStore = exports.pipeStateMutations = exports.declareStateUpdates = void 0;
const rxjs_1 = require("rxjs");
const compute_1 = require("./compute");
const storeEvents_1 = require("./storeEvents");
const storeMetadata_1 = require("./storeMetadata");
const utils_1 = require("./utils");
let STORE_SEQ_NUMBER = 0;
function declareStateUpdates(stateExample, updates) {
    if (updates) {
        return updates;
    }
    return (updates) => updates;
}
exports.declareStateUpdates = declareStateUpdates;
/**
 * Returns a mutation which applies all provided mutations for a state.
 *
 * You can use this helper to apply multiple changes at the same time.
 */
function pipeStateMutations(mutations) {
    return (state) => mutations.reduce((nextState, mutation) => mutation(nextState), state);
}
exports.pipeStateMutations = pipeStateMutations;
const NOTIFY_QUEUE = new Set();
/**
 * Creates the state store.
 *
 * @param initialState Initial state
 * @param options Parameters for the store
 */
function createStore(initialState, options) {
    var _a;
    const stateComparator = (_a = options === null || options === void 0 ? void 0 : options.comparator) !== null && _a !== void 0 ? _a : utils_1.DEFAULT_COMPARATOR;
    let currentState = initialState;
    const subscribers = new Set();
    const store = {
        id: ++STORE_SEQ_NUMBER,
        name: options === null || options === void 0 ? void 0 : options.name,
        value$: new rxjs_1.Observable((subscriber) => {
            subscribers.add(subscriber);
            subscriber.next(currentState);
            return () => subscribers.delete(subscriber);
        }),
        get() {
            return currentState;
        },
        query(selector) {
            return selector
                ? (0, compute_1.compute)((get) => selector(get(store)), [store])
                : store;
        },
        set(nextState) {
            update(() => nextState);
        },
        update,
        notify() {
            const pinnedState = currentState;
            subscribers.forEach((subscriber) => subscriber.next(pinnedState));
        },
        destroy() {
            var _a;
            subscribers.forEach((subscriber) => subscriber.complete());
            subscribers.clear();
            storeEvents_1.STORE_EVENT_BUS.next({ type: 'destroyed', store });
            (_a = options === null || options === void 0 ? void 0 : options.onDestroy) === null || _a === void 0 ? void 0 : _a.call(options);
        },
    };
    function update(mutation) {
        const prevState = currentState;
        const nextState = (0, utils_1.isReadonlyArray)(mutation)
            ? applyMutations(currentState, mutation)
            : applyMutation(currentState, mutation);
        if (stateComparator(prevState, nextState)) {
            return;
        }
        (0, compute_1.nextStoreVersion)();
        currentState = nextState;
        scheduleNotify(store);
        storeEvents_1.STORE_EVENT_BUS.next({
            type: 'updated',
            store,
            nextState,
            prevState,
        });
    }
    function applyMutations(state, mutations) {
        let nextState = state;
        for (let i = 0; i < mutations.length; i++) {
            const mutation = mutations[i];
            if (mutation) {
                nextState = applyMutation(nextState, mutation);
            }
        }
        return nextState;
    }
    function applyMutation(state, mutation) {
        const nextState = mutation(state);
        storeEvents_1.STORE_EVENT_BUS.next({
            type: 'mutation',
            store,
            mutation,
            nextState,
            prevState: state,
        });
        return nextState;
    }
    if (options === null || options === void 0 ? void 0 : options.internal) {
        (0, storeMetadata_1.setInternalStoreFlag)(store);
    }
    storeEvents_1.STORE_EVENT_BUS.next({ type: 'created', store });
    return store;
}
exports.createStore = createStore;
function scheduleNotify(store) {
    if (NOTIFY_QUEUE.add(store).size === 1) {
        Promise.resolve().then(executeNotifyQueue);
    }
}
function executeNotifyQueue() {
    const queue = [...NOTIFY_QUEUE];
    NOTIFY_QUEUE.clear();
    queue.forEach((store) => store.notify());
}
/** Creates StateUpdates for updating the store by provided state mutations */
function createStoreUpdates(storeUpdate, stateUpdates) {
    const updates = {};
    Object.entries(stateUpdates).forEach(([key, mutationFactory]) => {
        updates[key] = (...args) => {
            const mutation = mutationFactory(...args);
            (0, storeMetadata_1.setStateMutationName)(mutation, key);
            storeUpdate(mutation);
        };
    });
    return updates;
}
exports.createStoreUpdates = createStoreUpdates;
/** Creates a proxy for the store with "updates" to change a state by provided mutations */
function withStoreUpdates(store, updates) {
    const storeUpdates = createStoreUpdates(store.update, updates);
    return { ...store, updates: storeUpdates };
}
exports.withStoreUpdates = withStoreUpdates;
//# sourceMappingURL=store.js.map