import { Observer } from 'rxjs';
import { Query } from './query';
type Comparator<T> = (a: T, b: T) => boolean;
/**
 * This function returns a current value of a provided Query and registers it as a dependency for computation.
 */
export type ComputationResolver = {
    <T>(query: Query<T>): T;
    <T, R>(query: Query<T>, selector: (value: T) => R): R;
};
/**
 * This function calculates a value from external sources or resolved sources by a provided "resolver".
 */
export type Computation<T> = (resolver: ComputationResolver) => T;
/**
 * Options for "compute()" function
 */
export type ComputationOptions<T> = {
    /** A custom comparator to differ complex values */
    comparator?: Comparator<T>;
    /** Explicitly dependencies for refreshing calculations */
    dependencies?: Query<unknown>[];
};
/**
 * Creates a computable query which calculates its values by provided "computation" function and dependencies.
 *
 * Rules of "Computation" function:
 * - it must have no side effects
 * - it recalculates only when specified dependencies are updated
 * - its "formula" may have other sources of values, however they don't trigger updates
 *
 * "Computation" function provides a resolver for using a dependency withing a calculation expression.
 *
 * Dependency can be declared explicitly as an array by the second argument. It has the following advantages:
 * - Faster dependency subscription
 * - Ability to specify extra queries or observables as dependencies
 *
 * A custom value comparator can be specified by "options" object as the second argument.
 * It helps to decide if a new value differs from a previous one in complex cases.
 *
 * @example
 * ```ts
 * const greeting = createStore('Hello');
 * const username = createStore('World');
 *
 * // Dependency are implicitly resolved
 * const message = compute((get) => get(greeting) + ' ' + get(username) + '!');
 *
 * // Dependency declared explicitly
 * const messageUppercase = compute(() => message.get().toUpperCase(), [message]);
 *
 * expect(message.get()).toBe('Hello World!');
 * expect(messageUppercase.get()).toBe('HELLO WORLD!');
 * ```
 */
export declare const compute: {
    <T>(computation: Computation<T>, dependencies?: Query<unknown>[]): Query<T>;
    <T>(computation: Computation<T>, options?: ComputationOptions<T>): Query<T>;
};
type ValueRef<T> = {
    value: T;
    params?: Array<unknown>;
    version?: number;
};
/** @internal */
export type Node<T> = {
    computation: Computation<T>;
    comparator?: (a: T, b: T) => boolean;
    hot: boolean;
    version?: number;
    valueRef?: ValueRef<T>;
    dependencies?: Query<unknown>[];
    depsSubscriptions?: (() => void)[];
    observers?: Observer<T>[];
    treeObserverCount: number;
    parents?: Node<any>[];
    children?: Node<any>[];
};
/** @internal */
export declare function createComputationNode<T>(computation: Computation<T>, options?: ComputationOptions<T>): Node<T>;
/** @internal */
export declare function createComputationQuery<T>(node: Node<T>): Query<T>;
/** @internal */
export declare function getComputationNode<T>(query: Query<T>): Node<T> | undefined;
/** @internal */
export declare function nextVersion(currentValue: number): number;
/** @internal */
export declare function nextNodeVersion(): void;
/** @internal */
export declare function nextStoreVersion(): void;
export declare function getQueryValue<T>(node: Node<T>): T;
export declare function addValueObserver<T>(node: Node<T>, observer: Observer<T>): void;
export declare function removeValueObserver<T>(node: Node<T>, observer: Observer<T>): void;
export declare function addChildNode(parent: Node<any>, child: Node<any>): void;
export declare function makeColdNode<T>(node: Node<T>): void;
export declare function onSourceError<T>(node: Node<T>, error: any): void;
export declare function onSourceComplete<T>(node: Node<T>): void;
export declare function recompute<T>(node: Node<T>): void;
export declare function calculateValue<T>(node: Node<T>): void;
export {};
