import { Observable } from 'rxjs';
import { compute, nextStoreVersion } from './compute';
import { STORE_EVENT_BUS } from './storeEvents';
import { setInternalStoreFlag, setStateMutationName } from './storeMetadata';
import { DEFAULT_COMPARATOR, isReadonlyArray } from './utils';
let STORE_SEQ_NUMBER = 0;
export function declareStateUpdates(stateExample, updates) {
    if (updates) {
        return updates;
    }
    return (updates) => updates;
}
/**
 * Returns a mutation which applies all provided mutations for a state.
 *
 * You can use this helper to apply multiple changes at the same time.
 */
export function pipeStateMutations(mutations) {
    return (state) => mutations.reduce((nextState, mutation) => mutation(nextState), state);
}
const NOTIFY_QUEUE = new Set();
/**
 * Creates the state store.
 *
 * @param initialState Initial state
 * @param options Parameters for the store
 */
export function createStore(initialState, options) {
    var _a;
    const stateComparator = (_a = options === null || options === void 0 ? void 0 : options.comparator) !== null && _a !== void 0 ? _a : DEFAULT_COMPARATOR;
    let currentState = initialState;
    const subscribers = new Set();
    const store = {
        id: ++STORE_SEQ_NUMBER,
        name: options === null || options === void 0 ? void 0 : options.name,
        value$: new Observable((subscriber) => {
            subscribers.add(subscriber);
            subscriber.next(currentState);
            return () => subscribers.delete(subscriber);
        }),
        get() {
            return currentState;
        },
        query(selector) {
            return selector
                ? compute((get) => selector(get(store)), [store])
                : store;
        },
        set(nextState) {
            update(() => nextState);
        },
        update,
        notify() {
            const pinnedState = currentState;
            subscribers.forEach((subscriber) => subscriber.next(pinnedState));
        },
        destroy() {
            var _a;
            subscribers.forEach((subscriber) => subscriber.complete());
            subscribers.clear();
            STORE_EVENT_BUS.next({ type: 'destroyed', store });
            (_a = options === null || options === void 0 ? void 0 : options.onDestroy) === null || _a === void 0 ? void 0 : _a.call(options);
        },
    };
    function update(mutation) {
        const prevState = currentState;
        const nextState = isReadonlyArray(mutation)
            ? applyMutations(currentState, mutation)
            : applyMutation(currentState, mutation);
        if (stateComparator(prevState, nextState)) {
            return;
        }
        nextStoreVersion();
        currentState = nextState;
        scheduleNotify(store);
        STORE_EVENT_BUS.next({
            type: 'updated',
            store,
            nextState,
            prevState,
        });
    }
    function applyMutations(state, mutations) {
        let nextState = state;
        for (let i = 0; i < mutations.length; i++) {
            const mutation = mutations[i];
            if (mutation) {
                nextState = applyMutation(nextState, mutation);
            }
        }
        return nextState;
    }
    function applyMutation(state, mutation) {
        const nextState = mutation(state);
        STORE_EVENT_BUS.next({
            type: 'mutation',
            store,
            mutation,
            nextState,
            prevState: state,
        });
        return nextState;
    }
    if (options === null || options === void 0 ? void 0 : options.internal) {
        setInternalStoreFlag(store);
    }
    STORE_EVENT_BUS.next({ type: 'created', store });
    return store;
}
function scheduleNotify(store) {
    if (NOTIFY_QUEUE.add(store).size === 1) {
        Promise.resolve().then(executeNotifyQueue);
    }
}
function executeNotifyQueue() {
    const queue = [...NOTIFY_QUEUE];
    NOTIFY_QUEUE.clear();
    queue.forEach((store) => store.notify());
}
/** Creates StateUpdates for updating the store by provided state mutations */
export function createStoreUpdates(storeUpdate, stateUpdates) {
    const updates = {};
    Object.entries(stateUpdates).forEach(([key, mutationFactory]) => {
        updates[key] = (...args) => {
            const mutation = mutationFactory(...args);
            setStateMutationName(mutation, key);
            storeUpdate(mutation);
        };
    });
    return updates;
}
/** Creates a proxy for the store with "updates" to change a state by provided mutations */
export function withStoreUpdates(store, updates) {
    const storeUpdates = createStoreUpdates(store.update, updates);
    return { ...store, updates: storeUpdates };
}
//# sourceMappingURL=store.js.map